import sys
import webbrowser
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import parse_qsl
import requests
import secrets
import random
import logging
from urllib.parse import urlparse, urlencode

from pydeck_carto.carto_auth.errors import CredentialsError

logger = logging.getLogger(__name__)


def get_host_port(netloc):
    if ":" in netloc:
        host, port = netloc.split(":", 1)
        port = int(port)
    else:
        host = netloc
        port = None

    return host, port


class CartoPKCE:
    """Implements PKCE Authorization Flow for client apps"""

    OAUTH_AUTHORIZE_URL = "https://auth.carto.com/authorize"
    OAUTH_TOKEN_URL = "https://auth.carto.com/oauth/token"
    CLIENT_ID = "AtxvHDeuXlR8XPfF2nj2Uv2I29pvmCxu"
    REDIRECT_URI = "http://localhost:10000/callback/"
    REDIRECT_URI_CLI = "https://callbacks.carto.com/token"

    def __init__(
        self,
        open_browser=True,
    ):
        """
        Creates PKCE Auth flow.

        Parameters:
             * open_browser: Optional, whether the web browser should be opened to
                             authorize a user
        """
        self._session = requests.Session()
        self.client_id = self.CLIENT_ID
        self.redirect_uri = self.REDIRECT_URI if open_browser else self.REDIRECT_URI_CLI

        self._code_challenge_method = "S256"
        self.code_verifier = None
        self.code_challenge = None
        self.authorization_code = None

        using_google_colab = "google.colab" in sys.modules
        self.open_browser = False if using_google_colab else open_browser

    @staticmethod
    def _get_code_verifier():
        """Auth PCKE code verifier"""
        length = random.randint(33, 96)
        return secrets.token_urlsafe(length)

    def _get_code_challenge(self):
        """Auth PCKE code challenge"""
        import base64
        import hashlib

        code_challenge_digest = hashlib.sha256(
            self.code_verifier.encode("utf-8")
        ).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge_digest).decode("utf-8")
        return code_challenge.replace("=", "")

    def _open_auth_url(self, state=None):
        auth_url = self.get_authorize_url(state)
        try:
            opened = webbrowser.open_new(auth_url)
            if opened:
                logger.info("Opened %s in your browser", auth_url)
            else:
                raise webbrowser.Error()
        except webbrowser.Error:
            self.redirect_uri = self.REDIRECT_URI_CLI
            auth_url = self.get_authorize_url(state)
            logger.error("Please navigate here: %s", auth_url)

    def get_authorize_url(self, state=None):
        """Gets the URL to use to authorize this app"""
        if not self.code_challenge:
            self.get_pkce_handshake_parameters()
        payload = {
            "client_id": self.client_id,
            "response_type": "code",
            "scope": None,
            "audience": "carto-cloud-native-api",
            "redirect_uri": self.redirect_uri,
            "code_challenge_method": self._code_challenge_method,
            "code_challenge": self.code_challenge,
        }
        if state is not None:
            payload["state"] = state
        urlparams = urlencode(payload)
        return "%s?%s" % (self.OAUTH_AUTHORIZE_URL, urlparams)

    def get_auth_response(self, open_browser=None):
        logger.info(
            "User authentication requires interaction with your "
            "web browser. Once you enter your credentials and "
            "give authorization, you will be redirected to "
            "a url.  Paste that url you were directed to to "
            "complete the authorization."
        )

        redirect_info = urlparse(self.redirect_uri)
        redirect_host, redirect_port = get_host_port(redirect_info.netloc)

        if open_browser is None:
            open_browser = self.open_browser

        if (
            open_browser
            and redirect_host in ("127.0.0.1", "localhost")
            and redirect_info.scheme == "http"
        ):
            return self._get_auth_response_local_server(redirect_port)
        return self._get_auth_response_interactive(open_browser=open_browser)

    def _get_auth_response_local_server(self, redirect_port):
        server = start_local_http_server(redirect_port)
        self._open_auth_url()
        server.handle_request()

        if server.auth_code is not None:
            return server.auth_code
        elif server.error is not None:
            raise CredentialsError(
                "Received error from OAuth server: {}".format(server.error)
            )
        else:
            raise CredentialsError(
                "Server listening on localhost has not been accessed"
            )

    def _input(self, prompt):
        return input(prompt)

    def _get_auth_response_interactive(self, open_browser=False):
        if open_browser or self.open_browser:
            self._open_auth_url()
            prompt = "Enter the URL you were redirected to: "
        else:
            url = self.get_authorize_url()
            prompt = "Go to the following URL:\n{}\nEnter the Access Code: ".format(url)
        response = self._input(prompt)
        if "code=" in response:
            state, code = self.parse_auth_response_url(response)
        else:
            # just in case the user just copy and paste the access code from the button
            code = response
        return code

    def get_authorization_code(self, response=None):
        if response:
            return self.parse_response_code(response)
        return self.get_auth_response()

    def get_pkce_handshake_parameters(self):
        self.code_verifier = self._get_code_verifier()
        self.code_challenge = self._get_code_challenge()

    def get_token_info(self, code=None):
        """Gets the access token for the carto API using the authorization code"""
        if self.code_verifier is None or self.code_challenge is None:
            self.get_pkce_handshake_parameters()

        payload = {
            "client_id": self.client_id,
            "grant_type": "authorization_code",
            "audience": "carto-cloud-native-api",
            "code": code or self.get_authorization_code(),
            "redirect_uri": self.redirect_uri,
            "code_verifier": self.code_verifier,
        }

        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        logger.debug(
            "sending POST request to %s with Headers: %s and Body: %r",
            self.OAUTH_TOKEN_URL,
            headers,
            payload,
        )

        try:
            response = self._session.post(
                self.OAUTH_TOKEN_URL,
                data=payload,
                headers=headers,
                verify=True,
            )
            response.raise_for_status()
            token_info = response.json()
            return token_info
        except requests.exceptions.HTTPError as http_error:
            raise CredentialsError.from_http_error(http_error)

    def parse_response_code(self, url):
        """Parse the response code in the given response url"""
        _, code = self.parse_auth_response_url(url)
        if code is None:
            return url
        else:
            return code

    @staticmethod
    def parse_auth_response_url(url):
        query_s = urlparse(url).query
        form = dict(parse_qsl(query_s))
        if "error" in form:
            raise CredentialsError(f"Received error from auth server: {form['error']}")
        return tuple(form.get(param) for param in ["state", "code"])


class RequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.server.auth_code = self.server.error = None
        try:
            state, auth_code = CartoPKCE.parse_auth_response_url(self.path)
            self.server.state = state
            self.server.auth_code = auth_code
        except Exception as error:
            self.server.error = error

        self.send_response(200)
        self.send_header("Content-Type", "text/html")
        self.end_headers()

        if self.server.auth_code:
            callback_html = """<script>
            window.location.href = "https://callbacks.carto.com/success";
            </script>"""
        else:
            callback_html = """<script>
            window.location.href = "https://callbacks.carto.com/error";
            </script>"""

        self._write(callback_html)

    def _write(self, text):
        return self.wfile.write(text.encode("utf-8"))

    def log_message(self, format, *args):
        return


def start_local_http_server(port, handler=RequestHandler):
    server = HTTPServer(("localhost", port), handler)
    server.allow_reuse_address = True
    server.auth_code = None
    server.auth_token_form = None
    server.error = None
    return server



import os

import pydeck as pdk
import pytest
from pydeck_carto.carto_auth.auth import CartoAuth
from pydeck_carto.carto_auth.layer_validator import is_valid_carto_layer
from pydeck_carto.layer import MapType, CartoConnection, GeoColumnType


@pytest.fixture
def carto_dummy_auth(requests_mock):
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX.payload.signature",
            "scope": "read:tokens write:tokens read:imports write:imports "
            "read:connections write:connections",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )
    filepath = "fixtures/mock_credentials.json"
    fullpath = os.path.join(os.path.dirname(__file__), filepath)
    return CartoAuth.from_file(filepath=fullpath, use_cache=False)


@pytest.mark.parametrize(
    "layer_type, must_error",
    (
        (MapType.QUERY, False),
        (MapType.TABLE, False),
        (MapType.TILESET, False),
        ("query", True),
        ("table", True),
        (pdk.types.String("my_own_type"), True),
    ),
)
def test_correct_carto_layer(carto_dummy_auth, layer_type, must_error):
    layer = pdk.Layer(
        "CartoLayer",
        data="SELECT geom, name FROM carto-demo-data.demo_tables.airports",
        type_=layer_type,
        connection=CartoConnection.CARTO_DW,
        credentials=carto_dummy_auth.get_layer_credentials(),
        get_fill_color=[238, 77, 90],
        point_radius_min_pixels=2.5,
        pickable=True,
    )
    if must_error:
        with pytest.raises(ValueError):
            assert is_valid_carto_layer(layer, carto_dummy_auth) is True
    else:
        assert is_valid_carto_layer(layer, carto_dummy_auth) is True


def test_check_credential_token():
    correct_token = "header.payload.signature"
    ca = CartoAuth(access_token=correct_token, expires_in=7200)
    correct_layer = pdk.Layer(
        "CartoLayer",
        data="SELECT geom, name FROM carto-demo-data.demo_tables.airports",
        type_=MapType.QUERY,
        connection=CartoConnection.CARTO_DW,
        credentials=ca.get_layer_credentials(),
        get_fill_color=[238, 77, 90],
        point_radius_min_pixels=2.5,
        pickable=True,
    )
    assert is_valid_carto_layer(correct_layer, ca) is True

    with pytest.raises(ValueError):
        incorrect_token = "pepito..signature"
        ca = CartoAuth(access_token=incorrect_token, expires_in=7200)
        layer = pdk.Layer(
            "CartoLayer",
            data="SELECT geom, name FROM carto-demo-data.demo_tables.airports",
            type_=MapType.QUERY,
            connection=CartoConnection.CARTO_DW,
            credentials=ca.get_layer_credentials(),
            get_fill_color=[238, 77, 90],
            point_radius_min_pixels=2.5,
            pickable=True,
        )
        _ = is_valid_carto_layer(layer, ca)


@pytest.mark.parametrize(
    "layer_geo_column, must_error",
    (
        (GeoColumnType.H3, False),
        (GeoColumnType.QUADBIN, False),
        (pdk.types.String("randomColumn"), False),
        ("RandomName", True),
    ),
)
def test_geo_column_layer_validity(carto_dummy_auth, layer_geo_column, must_error):
    layer = pdk.Layer(
        "CartoLayer",
        data="SELECT geom, name FROM carto-demo-data.demo_tables.airports",
        type_=MapType.QUERY,
        connection=CartoConnection.CARTO_DW,
        credentials=carto_dummy_auth.get_layer_credentials(),
        geo_column=layer_geo_column,
        get_fill_color=[238, 77, 90],
        point_radius_min_pixels=2.5,
        pickable=True,
    )
    if must_error:
        with pytest.raises(ValueError):
            assert is_valid_carto_layer(layer, carto_dummy_auth) is False
    else:
        assert is_valid_carto_layer(layer, carto_dummy_auth) is True


def test_layers_missing_attrs(carto_dummy_auth):
    for mandatory_attr in ["data", "type_", "connection", "credentials"]:
        params = dict(
            data="SELECT geom, name FROM carto-demo-data.demo_tables.airports",
            type_=MapType.QUERY,
            connection=CartoConnection.CARTO_DW,
            credentials=carto_dummy_auth.get_layer_credentials(),
        )
        del params[mandatory_attr]

        layer = pdk.Layer("CartoLayer", **params)
        with pytest.raises(ValueError):
            assert is_valid_carto_layer(layer, carto_dummy_auth) is False


def test_list_connections(requests_mock, carto_dummy_auth):
    requests_mock.get(
        "https://api.carto.com/v3/connections",
        json=[
            {
                "name": "my_own_connection",
                "provider_id": "bigquery",
                "config": {
                    "type": "oauth_token",
                    "credentials": {
                        "picture": "picture-url",
                        "expires_in": 1661331110615,
                        "client_email": "super-email@exqample.com",
                    },
                    "billing_project": "my-billing",
                },
                "user_id": "google-oauth2|12345",
                "account_id": "ac_custom",
                "created_at": "2022-08-24T07:52:22.169Z",
                "updated_at": "2022-08-24T10:50:56.468Z",
                "id": "8b9d630e-ddb5-4796-8a16-d2323072fbbb",
                "spatial_extension": False,
                "privacy": "private",
                "owner_picture": "https://pictures-owner.jpg",
                "mapcount": 1,
                "carto_dw": False,
            },
        ],
    )

    conns = carto_dummy_auth.list_connections()
    assert isinstance(conns, list)

    random_str_connection = "myrandom_connnection"
    assert random_str_connection not in conns

    first_connection = "my_own_connection"
    assert first_connection in conns



"""

Testing the credentials object with different features
Create credentials
* from filesystem
* from arguments
* from auth access_token and expired_in

ask for auth from script

dump and red token credentials into a hidden file

Check if the token has been expired, refresh token function
"""
import datetime
import os

from pydeck_carto import CartoAuth, CredentialsError


def test_credentials_from_file(requests_mock):
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
            "scope": "read:tokens write:tokens read:imports write:imports "
            "read:connections write:connections",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )

    filepath = "fixtures/mock_credentials.json"
    fullpath = os.path.join(os.path.dirname(__file__), filepath)
    ca = CartoAuth.from_file(filepath=fullpath)
    assert ca.client_id == "1234"
    assert ca.client_secret == "1234567890"
    assert ca.api_base_url == "https://api.carto.com"

    assert ca.token_expired() is True

    creds = ca.get_layer_credentials()
    assert creds["apiVersion"] == "v3"
    assert creds["apiBaseUrl"] == ca.api_base_url
    assert len(creds["accessToken"]) > 0

    access_token = ca._get_token()
    assert len(access_token) > 0
    current_utc_ts = datetime.datetime.utcnow().timestamp()
    assert ca.expiration_ts > current_utc_ts

    assert ca.token_expired() is False


def test_credentials_from_parameters(requests_mock):
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
            "scope": "read:tokens write:tokens read:imports write:imports "
            "read:connections write:connections",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )

    ca = CartoAuth(client_id="1234", client_secret="1234567890")
    assert ca.client_id == "1234"
    assert ca.client_secret == "1234567890"
    assert ca.api_base_url == "https://gcp-us-east1.api.carto.com"

    creds = ca.get_layer_credentials()
    assert creds["apiVersion"] == "v3"
    assert creds["apiBaseUrl"] == ca.api_base_url
    assert len(creds["accessToken"]) > 0


def test_handle_file_token_cached_on_file():
    cache_filepath = "fixtures/.carto_token.json"
    fullpath = os.path.join(os.path.dirname(__file__), cache_filepath)
    ca = CartoAuth(cache_filepath=fullpath)

    saved_token = "testAccessTokenlkjsdofiuqwelrkjas908d7"  # encoded on the file
    assert saved_token == ca._get_token()

    expected_expiration_ts = 32503676400  # encoded on the file
    assert ca.expiration_ts == expected_expiration_ts
    assert ca.token_expired() is False


def test_handle_file_token_cached_expired_on_file():
    cache_filepath = "fixtures/.carto_token_expired.json"
    fullpath = os.path.join(os.path.dirname(__file__), cache_filepath)
    try:
        _ = CartoAuth(cache_filepath=fullpath)
        assert 1 == 2, "Saved expired token returned instead of an error"
    except CredentialsError:
        assert 1 == 1


def test_handle_auth_tokens():
    access_token = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6InUzWjRJOVBhUVZ2RjA2MWVaZHlfNCJ9.eyJpc3MiOiJodHRwczovL29yYW1pcmV6LWF1dGguZXUuYXV0aDAuY29tLyIsInN1YiI6Imdvb2dsZS1vYXV0aDJ8MTEwNjc5NzE5MDU5ODk3NjU5Njc4IiwiYXVkIjoiYXBpL2ZydWl0IiwiaWF0IjoxNjYxNzU4MTE1LCJleHAiOjE2NjE3NjUzMTUsImF6cCI6InN0TExYYnF1b2o2dnVOUDNCYUZaeEhmNE1UcFpLdWdhIiwic2NvcGUiOiJ3cml0ZTp1c3VhbC1mcnVpdCJ9.E5DbjwEOlbm37GkJPqkvh2ZZjlxtvERItd4loiWNcUXrzGQ34PBiC-kESeIFk_7AicldWxxGmIBZKycBflwoDNkR3CRchgW2CW4lPK7SSNAVR5iTwPPAHAhLwgwgKNmldOv5Wq7sDhf7Rc0JAQAdBqaTHXQrae57LNlnxVN--uJdq3oc4LYE3NDvQJIHlWaPDPsE6IgTOdQuS5bY867Ux3u3iLxEJpOevsce0d8l2or3se6GTX8rTb2Ip8rTkIrll0qzl-uwgTy0AoD-HM748W_FA1ScdTboilCzko6cFsMIDzb-ou-5BQPkIp6GtsYZuwrSD_t6t_ZrP46ehoUd1A"  # noqa: E501
    expire_in = 7200
    ca = CartoAuth(access_token=access_token, expires_in=expire_in)

    creds = ca.get_layer_credentials()
    assert creds["apiVersion"] == "v3"
    assert creds["apiBaseUrl"] == ca.api_base_url
    assert creds["accessToken"] == access_token

    assert ca.token_expired() is False



from urllib.parse import urlparse, parse_qs

from pydeck_carto import CartoAuth
from pydeck_carto.carto_auth.oauth2 import CartoPKCE


def test_url_properly_created():
    c_pkce = CartoPKCE(open_browser=False)
    url = c_pkce.get_authorize_url()
    parsed_url = urlparse(url)
    parsed_qs = parse_qs(parsed_url.query)

    assert parsed_qs["code_challenge_method"] == ["S256"]
    assert parsed_qs["redirect_uri"] == ["https://callbacks.carto.com/token"]
    assert parsed_qs["response_type"] == ["code"]
    assert parsed_qs["client_id"][0].startswith("AtxvHDe")
    assert len(parsed_qs["code_challenge"]) > 0


def test_token_from_input_prompt(mocker, requests_mock):
    mocker.patch(
        "pydeck_carto.carto_auth.oauth2.CartoPKCE._input",
        return_value="carto.com/autorize?code=abcde",
    )
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
            "scope": "read:tokens write:tokens read:imports write:imports "
            "read:connections write:connections",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )

    c_pkce = CartoPKCE(open_browser=False)
    c_pkce.get_pkce_handshake_parameters()

    assert len(c_pkce.code_challenge) > 0
    assert len(c_pkce.code_verifier) > 0


def test_token_from_input_mocked(mocker, requests_mock):
    mocker.patch(
        "pydeck_carto.carto_auth.oauth2.CartoPKCE._input",
        return_value="carto.com/autorize?code=abcde",
    )
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
            "scope": "",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )

    c_pkce = CartoPKCE(open_browser=False)
    code = c_pkce.get_auth_response()
    assert code == "abcde"


def test_carto_auth_from_oauth(mocker, requests_mock):
    mocker.patch(
        "pydeck_carto.carto_auth.oauth2.CartoPKCE._input",
        return_value="carto.com/autorize?code=abcde",
    )
    requests_mock.post(
        "https://auth.carto.com/oauth/token",
        json={
            "access_token": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
            "scope": "",
            "expires_in": 86400,
            "token_type": "Bearer",
        },
    )

    carto_auth = CartoAuth.from_oauth(open_browser=False, use_cache=False)
    assert carto_auth.get_layer_credentials() == {
        "apiVersion": "v3",
        "apiBaseUrl": "https://gcp-us-east1.api.carto.com",
        "accessToken": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpX",
    }


def test_handle_auth_tokens():
    access_token = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6InUzWjRJOVBhUVZ2RjA2MWVaZHlfNCJ9.eyJpc3MiOiJodHRwczovL29yYW1pcmV6LWF1dGguZXUuYXV0aDAuY29tLyIsInN1YiI6Imdvb2dsZS1vYXV0aDJ8MTEwNjc5NzE5MDU5ODk3NjU5Njc4IiwiYXVkIjoiYXBpL2ZydWl0IiwiaWF0IjoxNjYxNzU4MTE1LCJleHAiOjE2NjE3NjUzMTUsImF6cCI6InN0TExYYnF1b2o2dnVOUDNCYUZaeEhmNE1UcFpLdWdhIiwic2NvcGUiOiJ3cml0ZTp1c3VhbC1mcnVpdCJ9.E5DbjwEOlbm37GkJPqkvh2ZZjlxtvERItd4loiWNcUXrzGQ34PBiC-kESeIFk_7AicldWxxGmIBZKycBflwoDNkR3CRchgW2CW4lPK7SSNAVR5iTwPPAHAhLwgwgKNmldOv5Wq7sDhf7Rc0JAQAdBqaTHXQrae57LNlnxVN--uJdq3oc4LYE3NDvQJIHlWaPDPsE6IgTOdQuS5bY867Ux3u3iLxEJpOevsce0d8l2or3se6GTX8rTb2Ip8rTkIrll0qzl-uwgTy0AoD-HM748W_FA1ScdTboilCzko6cFsMIDzb-ou-5BQPkIp6GtsYZuwrSD_t6t_ZrP46ehoUd1A"  # noqa: E501
    expire_in = 7200
    ca = CartoAuth(access_token=access_token, expires_in=expire_in)

    creds = ca.get_layer_credentials()
    assert creds["apiVersion"] == "v3"
    assert creds["apiBaseUrl"] == ca.api_base_url
    assert creds["accessToken"] == access_token

    assert ca.token_expired() is False



from requests import HTTPError


class CredentialsError(Exception):
    @staticmethod
    def from_http_error(http_error: HTTPError):
        response = http_error.response
        try:
            error_payload = response.json()
            error = error_payload.get("error")
            error_description = error_payload.get("error_description")
        except ValueError:
            error = response.text or None
            error_description = None

        return CredentialsError(
            f"error: {error}, " f"error_description: {error_description}"
        )



"""
Test the BigQuery client connector created from a CartoAuth object

Creation of the bq client
requesting:
* datasets = client.list_datasets()
* tables = client.list_tables('shared')

"""
import os

from pydeck_carto import CartoAuth
from google.cloud.bigquery import Client as GClient

dummy_response = {
    "projectId": "project-id-mock",
    "token": "token-mock",
}


def test_cw_credentials(requests_mock):
    requests_mock.get(
        "https://gcp-us-east1.api.carto.com/v3/connections/carto-dw/token",
        json=dummy_response,
    )
    cache_filepath = "fixtures/.carto_token.json"
    fullpath = os.path.join(os.path.dirname(__file__), cache_filepath)
    ca = CartoAuth(cache_filepath=fullpath)

    carto_dw_project_id, carto_dw_token = ca.get_carto_dw_credentials()
    assert carto_dw_project_id == "project-id-mock"
    assert carto_dw_token == "token-mock"


def test_big_query_client(requests_mock):
    requests_mock.get(
        "https://gcp-us-east1.api.carto.com/v3/connections/carto-dw/token",
        json=dummy_response,
    )
    cache_filepath = "fixtures/.carto_token.json"
    fullpath = os.path.join(os.path.dirname(__file__), cache_filepath)
    ca = CartoAuth(cache_filepath=fullpath)

    bq_client = ca.get_carto_dw_client()
    assert bq_client is not None

    assert hasattr(bq_client, "query")
    assert isinstance(bq_client, GClient)



from ._version import __version__
from .layer import register_carto_layer
from .carto_auth.auth import CartoAuth, CredentialsError
from .carto_auth.layer_validator import is_valid_carto_layer

__all__ = [
    "__version__",
    "register_carto_layer",
    "CartoAuth",
    "CredentialsError",
    "is_valid_carto_layer",
]



from pydeck import Layer, types
from pydeck_carto.carto_auth.auth import CartoAuth
from pydeck_carto.layer import MapType, CartoConnection


def validate_attr(layer, attr, param_type):
    if not hasattr(layer, attr):
        raise ValueError(f"Missing parameter {attr} on the layer")
    attr_value = getattr(layer, attr)
    if not attr_value:
        raise ValueError(f"Parameter {attr} not valid")
    if not isinstance(attr_value, param_type):
        raise ValueError(f"Parameter {attr} must be a {param_type}")


def is_valid_carto_layer(layer: Layer, carto_auth: CartoAuth):
    if layer.type != "CartoLayer":
        raise ValueError('The layer type must be "CartoLayer"')

    validate_attr(layer, "credentials", dict)
    validate_attr(layer, "data", str)
    validate_attr(layer, "type_", types.String)
    validate_attr(layer, "connection", types.String)

    if hasattr(layer, "geo_column") and layer.geo_column is not None:
        validate_attr(layer, "geo_column", types.String)

    if layer.type_ not in [MapType.QUERY, MapType.TABLE, MapType.TILESET]:
        raise ValueError("Invalid layer type selected")

    for attr in ("accessToken", "apiVersion", "apiBaseUrl"):
        if attr not in layer.credentials:
            raise ValueError(f"Missing {attr} from layer.credentials")

    token_parts = layer.credentials["accessToken"].split(".")
    if len(token_parts) != 3:
        raise ValueError("Access token with wrong format")
    else:
        for part, part_label in zip(token_parts, ["header", "payload", "signature"]):
            if not part:
                raise ValueError(f'Missing "{part_label}" of the token')

    if layer.connection != CartoConnection.CARTO_DW:
        available_connections = carto_auth.list_connections()
        if layer.connection not in available_connections:
            raise ValueError(
                f"Invalid connection selected {layer.connection},"
                f" choose one of {available_connections}"
            )

    if layer.type_ == MapType.QUERY:
        cnt_params_expected = layer.data.count("?")
        if cnt_params_expected > 0 and cnt_params_expected != len(
            layer.query_parameters
        ):
            raise ValueError(
                "Mismatch query_parameters and parameters expected on the query"
            )

    return True
